import re
from py2neo import Graph, Node, Relationship, NodeMatcher, ClientError
from collections import defaultdict
from .semper_mapping import classes_for_token


class GraphUtils():

    def __init__(self, graph):
        self.graph = graph

    def delete_graph_for_file(self, filename):
        cursor = self.graph.run(
            'MATCH(n) WHERE n.filename=$filename DETACH DELETE n', 
            parameters = {
                "filename": filename
            }
        )

    def delete_graph(self):
        cursor = self.graph.run( 'MATCH(n) DETACH DELETE n' )


    def get_teis_for_categories(self, categories=None, operator="AND"):
        """For given categories, this method searches database and returns TEI nodes
        that are of these categories.
        """
        if categories == None:
            categories = []
        elif not isinstance(categories, list):
            categories=[categories]

        wheres = ['"{}" IN tei.categories'.format(category) for category in categories]
        where = (" "+operator+" ").join(wheres)
        if not where:
            # if no categories are provided, make sure that at least only return
            # TEI documents that contain the categories attribute
            where = "EXISTS( tei['categories'])"

        cypher = """
            MATCH (tei :TEI)
            WHERE {}
            RETURN tei
        """
        cursor = self.graph.run(cypher.format(where))

        teis = []
        while cursor.forward():
            for tei in cursor.current:
                teis.append(tei)
        return teis

    def get_info_for_filename(self, filename):
        """Returns general information about a given filename
        """

        cypher = """
            MATCH (tei :TEI {filename:$filename})-[:HAS_CATEGORY]->(cat :category)-[:HAS_DESCRIPTION]->(catDesc :catDesc)
            OPTIONAL MATCH (title :title {filename:$filename})
            OPTIONAL MATCH (graphic :graphic {filename:$filename})
            OPTIONAL MATCH (:msItem)--(locus :locus {filename:$filename})
            OPTIONAL MATCH (graphic :graphic {filename: "20_Ms_229_40.xml"})
            WITH tei, title, graphic, locus, cat, collect(catDesc {.language, .description}) as catDescs
            RETURN tei, title, graphic, locus, collect(cat {.`xml:id`, descriptions:catDescs }) as categories
        """
        cursor = self.graph.run(
            cypher,
            parameters = {
                "filename": filename
            }
        )
        info = {}
        while cursor.forward():
            tei, title, graphic, locus, categories = cursor.current
            info['tei'] = dict(tei.items())
            info['graphic'] = dict(graphic.items())
            info['locus'] = locus.get('string')
            info['categories'] = categories
            info['title'] = title.get('string')

        return info


    def paragraphs_for_filename(self, filename, only_with_matching_zone=True):
        kwargs = { "filename": filename }

        cypher = ""
        if only_with_matching_zone:
            cypher = """
                MATCH (paragraph :p {filename:$filename})-[:facs]->(zone :zone)
                OPTIONAL MATCH (zone)<-[:CONTAINS]-(:surface)-[:CONTAINS]->(graphic :graphic)
                RETURN paragraph, zone, graphic
            """
        else:
            cypher = """
                MATCH (paragraph :p {filename:$filename})
                OPTIONAL MATCH (p)-[:facs]->(zone :zone)
                OPTIONAL MATCH (zone)<-[:CONTAINS]-(:surface)-[:CONTAINS]->(graphic :graphic)
                RETURN paragraph, zone, graphic
            """
        cursor = self.graph.run(
            cypher,
            parameters = {
                "filename": filename
            }
        )

        paragraphs = []
        while cursor.forward():
            paragraph, zone, graphic = cursor.current
            paragraph.zone = zone
            paragraph.graphic = graphic
            paragraphs.append(paragraph)

        return paragraphs


    def nodes_with_link_for_filename(self, filename):
        """for a given filename, this query returns all nodes
        that contain any of the link_attributes (see below)
        The goal is to create relationships from these nodes to their targets
        """

        kwargs = { "filename": filename }
        link_attributes = ["facs", "next", "prev", "target", "ref"]
        where = " OR ".join("EXISTS(n.{})".format(attr) for attr in link_attributes)
        cypher = """
            MATCH (n)
            WHERE 
        """
        cypher += where
        cypher += " RETURN n"

        nodes = []
        while cursor.forward():
            for nodes in cursor.current:
                nodes.append(nodes)

        return nodes


    def concatenation_exists(self, node):
        cypher = """
        MATCH (node1)-[r :CONCATENATED]->(node2)
        WHERE ID(node1) = $node_id
        RETURN r
        """
        cursor = self.graph.run(cypher, parameters={"node_id": node.identity})

        # return 1 if there exists a relation
        return cursor.forward()

    def get_categories(self):
        cypher = """
        MATCH (cat :category)
        RETURN cat.`xml:id` as id, cat.description
        """
        cats = []
        while cursor.forward():
            for cat in cursor.current:
                cats.append(cat)

        return cats


    def paras2dict(self, paras, words_concatenated=0, follow_metamarks=True):
        """Takes an array of paragraph tokens
        returns a dictionary:
        {
            "zones": {}
            "paragraphs": []
        }
        """

        # translate True/False values to 0 and 1
        if words_concatenated:
            words_concatenated = 1
        else:
            words_concatenated = 0

        paras_dict = {}
        zones = {}
        paras_dict["zones"] = zones
        paras_dict["paragraphs"] = []
        delspan_to = ""

        for para in paras:
            zones["#"+para.zone.get('xml:id')]= {
                "points": para.zone.get('points'),
                "rendition": para.zone.get('rendition'),
                "subtype": para.zone.get('subtype'),
            }
            is_add = False
            metamark_start = False
 
            current_classes = set()
            create_new_token_group = False
 
            current_para = para
 
            current_token_group = None
            elements = []
 
            paragraph_dict = {
                "facs": para.get('facs'),
                "xml:id": para.get('xml:id'),
                "_elements": elements
            }
            paras_dict["paragraphs"].append(paragraph_dict)
 
            string = ""
            for token in self.tokens_in_paragraph(
                para,
                concatenated=words_concatenated,
                follow_metamarks=follow_metamarks
            ):
                    
                new_facs = ""

                # handle end of an add element
                if is_add and not 'add' in token.labels:
                    # mark end of the addition: ⟩
                    elements.append({
                        "string": "⟩",
                        "classes": list(current_classes)
                    })
                    is_add = False

                # handle end of a delSpan element
                if 'anchor' in token.labels and delspan_to and token.get('anchor','') in delspan_to:
                    elements.append({
                        "string": "◀",
                        "classes": list(current_classes),
                    })
                    delspan_to=""

                # handle metamarks: the token has a different paragraph_id
                # create a new token group when this happens
                if not current_para.get('xml:id') == token.get('paragraph_id'):
                    current_para = self.get_paragraph(
                        filename=token.get('filename'),
                        xml_id = token.get('paragraph_id')
                    )
                    create_new_token_group = True
                    new_facs = current_para.get('facs')

                # check if anything changed
                new_classes = classes_for_token(token)
                if not new_classes == current_classes:
                    create_new_token_group = True
 
                # handle <add> elements: add ⟨here is added string⟩
                # prepend another element
                if 'add' in token.labels and not is_add:
                    # mark addition-start: ⟨
                    # put it in front of current strings
                    elements.append({
                        "string": "⟨",
                        "classes": list(current_classes)
                    })
                    is_add = True
 
                # handle beginning of a delSpan: prepend another element
                if 'delSpan' in token.labels and not delspan_to:
                    elements.append({
                        "string": "▶",
                        "classes": list(current_classes),
                    })
                    delspan_to = token.get('spanTo')


                # create a new token group: all strings will be appended
                # as long they have the same style
                if create_new_token_group or current_token_group is None:
                    current_classes = new_classes
                    current_token_group = {
                        "string": "",
                        "classes": list(current_classes),
                    }
                    if new_facs:
                        current_token_group["facs"] = new_facs
                        new_facs = None
                    create_new_token_group = False

                    # append the token group to our elements
                    elements.append(current_token_group)
 
                # handle linebreaks
                if 'lb' in token.labels:
                    current_token_group['string'] += " | "
                    continue

                # append the string, including whitespace
                current_token_group['string'] += token.get('string', '') + token.get('whitespace','')


            # is still in add: close it when paragraph ends
            if is_add:
                elements.append({
                    "string": "⟩",
                    "classes": list(current_classes)
                })
                is_add = False

        # still in delSpan: close it when all paragraphs are done:
        if delspan_to:
            elements.append({
                "string": "◀",
                "classes": list(current_classes),
            })
            delspan_to=""


        return paras_dict


    def tokens_in_paragraph(self, paragraph:Node, concatenated=0, follow_metamarks=False):
        """For a given paragraph, this method returns all nodes
        connected via a NEXT relationship. 
        If concatenated=1, it will return a concatenated version of the textpath.
        Returns all nodes in the found textpath.
        """

        cypher="""
        MATCH (para)-[:NEXT]->(t),
        textpath = shortestPath((t)-[:NEXT*]->(lt))
        WHERE ID(para)=$paragraph_id
        AND (para)-[:LAST]->(lt)
        AND ALL (
            rel IN relationships(textpath)
            WHERE (rel.concatenated IS NULL OR rel.concatenated = $concatenated)
        )
        RETURN nodes(textpath)
        """
        # NOTE: it is assumed that a path containing a concatenated (non-hyphened)
        # word will be always shorter than a path containing a hyphened word.
        # The non-concatenated textpath actually never has the relation-attribute 
        # rel.concatenated=0 (it is always NULL)

        cursor = self.graph.run(
            cypher, 
            parameters={
                "paragraph_id": paragraph.identity,
                "concatenated": concatenated
            }
        ) 

        tokens = []
        while cursor.forward():
            for entry in cursor.current:
                for token in entry:
                    if follow_metamarks \
                    and 'metamark' in token.labels \
                    and token.get('function', '') == 'Einfügung' \
                    and token.get('target', ''):
                        tei_file, anchor = token.get('target').split('#')
                        if not tei_file:
                            tei_file = paragraph.get('filename')
                            target_paragraph = self.get_paragraph(
                                filename = tei_file,
                                xml_id = anchor
                            )
                            if target_paragraph:
                                metamark_tokens = self.tokens_in_paragraph(
                                    paragraph = target_paragraph,
                                    concatenated = concatenated,
                                    follow_metamarks = follow_metamarks
                                )
                                tokens += metamark_tokens
                    else:
                        tokens.append(token)
        return tokens


    def get_paragraph(self, filename, xml_id):
        """Returns a paragraph node for filename and xml_id
        """

        parameters = { 
            "filename": filename,
            "xml_id": xml_id,
        }

        cypher = """
            MATCH (paragraph :p {filename:$filename, `xml:id`: $xml_id})
            OPTIONAL MATCH (p)-[:facs]->(zone :zone)
            RETURN paragraph, zone
        """
        cursor = self.graph.run(cypher, parameters = parameters)

        paragraph = None
        while cursor.forward(amount=2):
            paragraph, zone = cursor.current
            paragraph.zone = zone

        return paragraph
                


    def tokens_in_paragraphs(self,
        paragraphs,
        concatenate_words=0,
        concatenate_paragraphs=0,
        follow_metamarks=0
    ):
        """For a given list of paragraphs (nodes) this method tries to:
        - return concatenated words,
        - concatenate paragraphs, when they are connected via next and prev attributes
        - follow metamarks, when linked via target attribute
        - remove duplicate paragraphs i.e. paragraphs that have already been
          visited should not appear again.
        - this method is not recursive, i.e. search-depth is 1, which should be fine in 99% of the cases.
        """
        paragraph_visited = {}
        paragraph_dict = {}
        for paragraph in paragraphs:
            paragraph_dict[paragraph.identity] = paragraph 

        nodes = []

        for paragraph in paragraphs:
            if paragraph.identity in paragraph_visited:
                continue
            nodes.append(paragraph)
            tokens = self.tokens_in_paragraph(paragraph, concatenated=concatenate_words)
            for token in tokens:
                nodes.append(token)
                if 'metamark' in token.labels\
                and token.get('target', False):
                    tei_file, anchor = token.get('target').split('#')
                    if anchor in paragraph_dict:
                        metamark_tokens = self.tokens_in_paragraph(
                            paragraph_dict[anchor], 
                            concatenated=concatenate_words
                        )
                        nodes += metamark_nodes
                        paragraph_visited[anchor] = True 

        return


    def create_unhyphenated(self, tokens):
        """tokens=Array of all tokens in a paragraph, as returned
        by GraphUtils.tokens_in_paragraph(para). This procedures looks for linebreaks with type=hyph
        If found, it looks forward and backwards to find the hyphened wordparts
        It then concatenates the wordparts, creates a new Node and new Releations.
        """
        tx = self.graph.begin()

        for i, token in enumerate(tokens):
            j=0
            k=0
            if token.has_label('lb'):
                wordstart = None
                wordend = None
                if token.get('type') == 'hyph':
                    
                    # walk back and find a token which is a wordpart
                    # and not any punctuation or similar
                    j=i-1
                    while j>0:
                        #print("j = {}".format(j))
                        if tokens[j].has_label('token') \
                        and tokens[j]["string"] \
                        and not tokens[j]["is_punct"]: 
                            #print("TRY: "+str(tokens[j]))
                            wordstart = tokens[j]
                            break
                        else:
                            j = j-1

                    # walk forward and find a token
                    k=i+1
                    while k>0 and tokens[k]:
                        #print("k = {}".format(k))
                        if tokens[k].has_label('token'):
                            #print("TRY END: "+str(tokens[k]))
                            wordend = tokens[k]
                            break
                        else:
                            k = k+1
                          
                    concat_word = ''
                    if wordstart and wordend and not self.concatenation_exists(wordstart):
                        #print("---START: "+str(wordstart))
                        #print("---  END: "+str(wordend))
                        if any(
                            wordstart["string"].endswith(s) for s\
                            in ['-', '\N{NOT SIGN}', '\N{NON-BREAKING HYPHEN}']
                        ):
                            concat_word = wordstart["string"][:-1]
                        else:
                            concat_word = wordstart["string"]
                            
                        concat_word += wordend["string"]
                        
                        # create new concatenated token
                        # with blank as whitespace
                        labels = list(
                            set(
                                ['token', 'concatenated'] + list(wordstart.labels) + list(wordend.labels)
                            )
                        )
                        attrs = { 
                            "string"    : concat_word,
                            "whitespace": wordend["whitespace"],
                            "filename"  : wordstart["filename"],
                            "idno"      : wordstart["idno"],
                        }
                        concat_node = Node(
                            *labels,
                            **attrs
                        )
                        #print("+ {}".format(concat_word))
                        tx.create(concat_node)
                        
                        # create relations from hyphened wordpards to concatenated word
                        rs = Relationship(
                            wordstart,
                            "CONCATENATED",
                            concat_node
                        )
                        rs2 = Relationship(
                            wordend,
                            "CONCATENATED",
                            concat_node
                        )
                        tx.create(rs)                
                        tx.create(rs2)
                        
                        # create direct connection for non-hyphened version
                        # of the thext
                        if j >0:
                            before_wordstart = tokens[j-1]
                            rs3 = Relationship(
                                before_wordstart,
                                "NEXT",
                                concat_node,
                                concatenated=1
                            )
                            tx.create(rs3)
                            
                        if len(tokens) > k+1:
                            after_wordend = tokens[k+1]
                            rs4 = Relationship(
                                concat_node,
                                "NEXT",
                                after_wordend,
                                concatenated=1
                            )
                            tx.create(rs4)
        tx.commit()

    def get_next_prev(self, filename):
        cypher = """
        MATCH (tei :TEI {filename:$filename})
        RETURN tei
        """
        cursor = self.graph.run(cypher, parameters={"filename":filename})
        tei = cursor.evaluate()
        attrs = {}
        if tei:
            for attr in ['first','last','next','prev']:
                filename = tei.get(attr, None)
                if filename:
                    match = re.search(r'\d+_Ms_(?P<collection>\d+?)_(?P<page>.*?).xml', filename)
                    if match:
                        attrs[attr] = match.groupdict()['page']
                        attrs['collection'] = match.groupdict()['collection']
        return attrs


    def connect_teis(self, *filenames):
        """Connects the TEI nodes with NEXT and PREV, given the filenames
        This allows to quickly show the previous and next pages of a document.
        The following attributes are also added to the TEI nodes:
        - first
        - last
        - prev
        - next
        """
        if len(filenames)<2:
            raise ValueError("Please provide at least two filenames")

        for i in range(len(filenames)-1):
            cypher = """
            MATCH (tei1 :TEI)
            WHERE tei1.filename = $prev
            MATCH (tei2 :TEI)
            WHERE tei2.filename = $next
            MERGE (tei1)-[:NEXT]->(tei2)
            MERGE (tei2)-[:PREV]->(tei1)
            SET tei1.next = $next,
                tei1.first = $first,
                tei1.last = $last,
                tei2.prev = $prev,
                tei2.first = $first,
                tei2.last = $last

            """ 
            parameters = {
                "prev"  : filenames[i],
                "next"  : filenames[i+1],
                "first" : filenames[0],
                "last"  : filenames[-1],
            }
            cursor = self.graph.run(cypher, parameters=parameters)


    def update_teis(self, *filenames, **keyvals):
        """
        updates the given TEI nodes (identified by filename) with the given
        keyval-pairs
        """

        for filename in filenames:
            cypher = """
            MATCH (tei :TEI { filename:$filename })
            SET tei += $keyvals
            """ 
            parameters = {
                "filename": filename,
                "keyvals": keyvals
            }
            cursor = self.graph.run(cypher, parameters=parameters)


    def connect_to_categories(self, filename):
        """Connects the categories in the catRef element
        with the existing categories in the graph database.
        Returns the category nodes for the given document.
        """
        cypher = """
        // match the catRef element and extract its category-targets
        MATCH (cat :catRef)
        WHERE cat.filename = $filename
        AND cat['target'] IS NOT NULL
        WITH SPLIT(cat['target'], " ") as pointers
        UNWIND pointers as pointer
        WITH SPLIT(pointer, "#") AS target

        // match the corresponding category targets
        MATCH (c :category)
        WHERE c.`xml:id` = target[1]

        // create relationships between the categories and the TEI node
        MATCH (tei :TEI)
        WHERE tei.filename = $filename
        MERGE (tei)-[:HAS_CATEGORY]->(c)
        RETURN c
        """

        parameters = { "filename": filename }
        cursor = self.graph.run(cypher, parameters = parameters)
        categories = []
        while cursor.forward():
            for category in cursor.current:
                categories.append(category)
        return categories

    def connect_references(self, filename):
        """Connects reference strings elements <rs> to its register nodes
        """

        cypher = """
        // match the rs element and extract its ref
        MATCH (rs :rs)
        WHERE rs.filename = $filename
        AND rs['ref'] IS NOT NULL
        AND rs['type'] IS NOT NULL
        WITH SPLIT(rs['ref'], " ") as pointers,
        rs['type'] as rs_type
        UNWIND pointers as pointer
        WITH rs_type, SPLIT(pointer, "#") AS target
        MATCH (ref)
        WHERE rs_type in labels(ref)
        AND ref.filename = target[0]
        AND ref.`xml:id` = target[1]

        // match the corresponding references
        MATCH (c :)
        WHERE c.`xml:id` = target[1]

        """

    def link_inner_relationships(self, filename):
        """creates new relationships between nodes of a given filename,
        that contain one of the link attributes below which point to
        a node inside the same xml file. Ignores all :join nodes, because
        they are handled by the handle_join_elements method.

        Within a link attribute, pointers can be separated by whitespace
        https://www.tei-c.org/release/doc/tei-p5-doc/en/html/ref-att.global.linking.html

        This method should be only run once after a TEI document has
        been parsed.
        Does not return anything.
        """
        
        link_attributes = ["facs", "next", "prev", "target", "ref"]

        cypher = """
        MATCH (from_node)
        WHERE from_node.filename = $filename
        AND NOT from_node:join
        AND from_node[$link_attribute] IS NOT NULL
        WITH from_node, SPLIT(from_node[$link_attribute], " ") AS pointers
        UNWIND pointers as pointer
        WITH from_node, SPLIT(pointer, "#") AS target
        WITH from_node, target,
        CASE WHEN target[0] = "" 
            THEN $filename 
            ELSE target[0] 
            END 
        AS filename
        MERGE (to_node {filename: filename, `xml:id`:target[1]} )
        """

        for link_attribute in link_attributes:
            parameters = {
                "filename"      : filename,
                "link_attribute": link_attribute,
            }
            create_rel = " CREATE (from_node)-[r:{}]->(to_node)".format(
                link_attribute.lower()
            )
            cursor = self.graph.run(cypher+create_rel, parameters = parameters)


    def tei_for_categories(self, *categories):
        exists_categories = []
        for category in categories:
            exists_categories.append(
                'EXISTS ((tei)-[:HAS_CATEGORY]->(:category {`xml:id`: "{}"}))'.format(category)
            )

        cypher = "MATCH (tei :TEI) WHERE"
        cypher += " AND ".join(exists_categories)
        cypher += " RETURN tei"
        cursor = self.graph.run(cypher)

        teis = []
        while cursor.forward():
            for tei in cursor.current:
                teis.append(category)
        return teis


    def handle_join_elements(self, filename):
        """Searches for all <join> elements for a given filename.
        Creates relationships between the mentioned elements of the «target» attribute
        Does not return anything.
        TODO: modify this method to establish a connection between two tokens
        on separate pages:
        (ta :token {`xml:id`:"p.page1"})-[:NEXT]->(tb :token {``xml:id`:"p.page2"})
        And update all tokens so they contain the identfiers of both joined paragraphs
        """
        
        cypher = """
            MATCH (join_node :join)
            WHERE join_node.filename = $filename
            AND join_node['target'] IS NOT NULL
            WITH join_node, SPLIT(join_node['target'], " ") AS endpoints
            WITH join_node, SPLIT(endpoints[0], "#") AS source, SPLIT(endpoints[1], "#") AS target
            WITH join_node, source, target,
            CASE WHEN source[0] = "" THEN $filename ELSE source[0] END AS source_filename,
            CASE WHEN target[0] = "" THEN $filename ELSE target[0] END AS target_filename
            MERGE (from_node {filename: source_filename, `xml:id`:source[1]})
            MERGE (to_node {filename: target_filename, `xml:id`:target[1]})
            MERGE (from_node)-[r :JOIN]->(to_node)
        """

        parameters = { "filename": filename }
        cursor = self.graph.run(cypher, parameters = parameters)


    def search_for_terms(self, terms, scope="page", categories=None):

        def get_search_mode_term(term):
            mode = ""
            if term.startswith("*") and term.endswith("*"):
                term = re.sub(r'\*', '', term)
                mode = "CONTAINS"
            elif term.startswith("*"):
                term = re.sub(r'\*', '', term)
                mode = "ENDS WITH"
            elif term.endswith("*"):
                term = re.sub(r'\*', '', term)
                mode = "STARTS WITH"
            else:
                mode = "="
            return (mode, term)


        if not isinstance(terms, list):
            terms = [terms]

        cypher = """
            MATCH (n0:token)
            WHERE n0.string {} "{}"
        """.format(*get_search_mode_term(terms[0]) )

        where_categories = ""
        if categories is not None:
            if not isinstance(categories, list):
                categories = [categories]
            where_categories = [
                '"{}" IN tei.categories'.format(category) for category in categories
            ]

        if where_categories:
            where_categories_str = (" AND ").join(where_categories)
            cypher += """
                MATCH (tei :TEI)
                WHERE {}
                AND tei.filename = n0.filename
            """.format(where_categories_str)


        if len(terms) > 1:
            for i, term in enumerate(terms):
                # a second or third term must be found within the same scope
                # (e.g. paragraph, filename, document)
                if i == 0: continue

                cypher +="""
                MATCH (n{}:token)
                WHERE n{}.string {} "{}"
                """.format(i, i, *get_search_mode_term(term))

                if scope == "paragraph":
                    # within the same document-collection (idno)
                    cypher += " AND n{}.idno = n0.idno ".format(i)
                    cypher += " AND n{}.paragraph_id = n0.paragraph_id ".format(i)
                # within the same TEI documents
                elif scope == "page":
                    cypher += " AND n{}.filename = n0.filename ".format(i)
                # within the same collection
                else:
                    cypher += " AND n{}.idno = n0.idno ".format(i)

        cypher +="""
            MATCH(para:p)
            WHERE para.idno=n0.idno
        """

        if scope == "paragraph":
            cypher += " AND para.`xml:id`= n0.paragraph_id"

        elif scope == "page":
            cypher += " AND para.filename= n0.filename"

        cypher += " return DISTINCT para"
        print(cypher)

        cursor = self.graph.run(cypher)

        paras = [] 
        while cursor.forward():
            for para in cursor.current:
                tokens = self.tokens_in_paragraph(para, concatenated=1)
                paras.append({
                    "para": para,
                    "tokens": tokens
                })

        return paras

        
    def create_indexes(self):
        labels = "del add rendition zone p pb anchor biblStruct category facsimile handNote item person place quote scriptNote".split() 
        for label in labels:
            cypher = "CREATE INDEX {}_idx FOR (n:{}) ON (n.filename, n.`xml:id`)".format(
                label, label
            )
            try:
                cursor = self.graph.run(cypher)
                print("{}_idx index created".format(label))
            except ClientError:
                print("{}_idx index already created".format(label))


    def get_indexes(self):
        cypher = "CALL db.indexes()"
        cursor = self.graph.run(cypher)
        indexes = []
        while cursor.forward():
            indexes.append(cursor.current)

        return indexes




import sys
assert sys.version_info >= (3, 5)  # at least Python 3.5 or higher

import os
from bs4 import BeautifulSoup, Comment
from py2neo import Graph, Node, Relationship, NodeMatcher
import spacy
from lxml import etree
import re
from collections import defaultdict

#from .tei_elements import *
import keyword
import inspect

nlp = spacy.load('de_core_news_sm')

VERBOSE = 0
PARSE_STRING = 0
CREATE_NODE = 0
GET_CLASS_FOR_NODE = 0
CHANGE_BASE_CLASS = 0
WALK_CHILD_NODES = 0
PARENT_CHILD_RELATION = 0

def get_categories(tei_node):
    """searches the TEI document for catRef
    and looks it up in the taxonomy/category definition
    returns a list of categories
    """
    categories = []
    catref_node = tei_node.find('catRef')
    if not catref_node: return categories
    catref_node = tei_node.find('catRef')
    for catref in catref_node.attrs.get('target','').split():
        filename, xml_id = catref.split('#')
        cat = tei_node.find(attrs={'xml:id':xml_id}) 
        if cat:
            for catdesc in cat.find_all('catDesc'):
                categories.append(catdesc.string)
    return categories 


def import_categories(filename, tx):

    def extract_information(cats, filename):
        all_cats = []
        for cat in cats.find_all('category'):
            catDescs = []
            for catDesc in cat.find_all('catDesc'):
                catDescs.append({
                    "language": catDesc.get('xml:lang'),
                    "description": catDesc.text
                })

            all_cats.append({
                "label": "category",
                "attrs": {
                    "xml:id": cat.get('xml:id'),
                    "filename": filename
                },
                "catDescs": catDescs
            })
        return all_cats

    def create_nodes_from_cats(cats, tx):
        for cat in cats:
            cat_node = Node(
                cat['label'],
                **cat['attrs']
            )
            tx.create(cat_node)
            for catDesc in cat['catDescs']:
                desc_node = Node(
                    'catDesc',
                    **catDesc
                )
                tx.create(desc_node)
                rs = Relationship(
                    cat_node,
                    'HAS_DESCRIPTION',
                    desc_node,
                )
                tx.create(rs)


    parser = etree.XMLParser(dtd_validation=True, recover=True)
    tree = etree.parse(filename, parser)
    unicode_string = etree.tostring(tree.getroot(), encoding='unicode')
    catsoup = BeautifulSoup(unicode_string, 'lxml-xml')

    cats = extract_information(catsoup, os.path.basename(filename))
    create_nodes_from_cats(cats, tx)


def parse(filename, start_with_tag='TEI', idno=None):
    status = defaultdict(list)
    status['attrs'] = defaultdict(dict)
    status['handshift'] = {}
    status['links'] = defaultdict(dict)
    attrs = {} 
    filename_only = os.path.basename(filename)
    #attrs['filename'] = filename_only
    status['attrs']['filename'] = filename_only

    # 1. parse the XML using lxml and etree.
    # It will apply the DTD validation and replace strings like &stern_1; with
    # the correct unicode symbol, which is described in the DTD file.
    parser = etree.XMLParser(dtd_validation=True, recover=True)
    tree = etree.parse(filename, parser)
    # output of XML document as unicode string
    unicode_string = etree.tostring(tree.getroot(), encoding='unicode')

    # 2. parse the TEI XML file (the unicode string), using lxml-xml parser
    # which is not assuming html and thus does not automatically add html or remove body tags.
    soup = BeautifulSoup(unicode_string, 'lxml-xml')

        # find the primary key of the document and attach it to the soup
    if not idno:
        idno = soup.find('idno').string
    
    status['attrs']['idno'] = idno

    tei_node = soup.find(start_with_tag)
    if tei_node is None:
        raise ValueError("Did not find a tag named '{}'".format(start_with_tag))

    # find a matching class, create an object from it
    # and return it
    class_name = start_with_tag.lower()
    if class_name in globals():
        if GET_CLASS_FOR_NODE: print("  * found in globals(): {}".format(globals()[class_name]))
        cls = globals()[class_name]
        doc = cls(
            node=tei_node, node_name=tei_node.name, status=status,
            attrs=attrs
        )
        return doc, status, tei_node
    else:
        raise ValueError("no such class available: {}".format(class_name))


# TEI section
class tei():
    """This is the base class for all elements in a TEI document.
    """

    def __init_subclass__( 
        cls, 
        parent_child_relation='CONTAINS', child_parent_relation=None, 
        isleaf=False, 
        do_not_store_this_node=False,
        pass_label_to_children=None,
        unpack_elements=False, 
        create_empty_token_nodes=False,
        from_relation_attr = None,
        to_relation_attr = None,
        group_similar_siblings = None,
        pass_attributes_to_children=False,
        cache_object=False,
        **kwargs
    ):
        super().__init_subclass__(**kwargs)
        # Parent defines,  how the relation from Parent to Child is defined
        cls.parent_child_relation = parent_child_relation
        # Child defines, how the relation from Child to Parent is defined
        cls.child_parent_relation = child_parent_relation
        cls.isleaf = isleaf  # extract all text inside this element
        cls.unpack_elements = unpack_elements # ignore this node, just add the elements
        # if an empty <sic/> element is encountered, create a token node
        # even the element contains no tokens at all
        cls.create_empty_token_nodes = create_empty_token_nodes       # create additional nodes that contain whitespace
        cls.from_relation_attr = from_relation_attr
        cls.to_relation_attr   = to_relation_attr
        cls.group_similar_siblings = group_similar_siblings           # see 
        cls.pass_label_to_children = pass_label_to_children           # add additional label to graph-node
        cls.pass_attributes_to_children = pass_attributes_to_children # add attributes of xml-node to graph-node
        cls.cache_object=cache_object
        cls.do_not_store_this_node = do_not_store_this_node


    def get_class_for_node(self, node):
        # add an underscore to the node name in case it is a reserved python word
        # which can't be used as a class name
        node_name = node.name
        if node_name in keyword.kwlist:
            node_name += '_'

        # inspect all subclasses in a class (and their parents)
        # and try to find a matching class to the given 
        # node_name (aka tag, element) in the XML file
        # return the class as soon as it is found, using 
        # C3 mro (method resolution order)

        if GET_CLASS_FOR_NODE: print("**** looking here in get_class_for_node for: ", node_name)
        for cls in inspect.getmro(self.__class__):
            if cls.__name__ == 'object':
                continue
            if GET_CLASS_FOR_NODE: print("   - looking for members in class: {}".format(cls))

            for subclass_tuple in inspect.getmembers(cls, predicate=lambda cl:(inspect.isclass(cl))):
                if GET_CLASS_FOR_NODE: print("      - looking at tuple: {}".format(subclass_tuple))
                if node_name.lower() == subclass_tuple[0].lower():
                    return subclass_tuple[1]

        # still no matching class found. Try to find a base class
        if GET_CLASS_FOR_NODE: print("   - looking in globals()")
        if node_name.lower() in globals():
            if GET_CLASS_FOR_NODE: print("  * found in globals(): {}".format(globals()[node_name.lower()]))
            return globals()[node_name.lower()]
        else:
            if GET_CLASS_FOR_NODE:
                print("XXXXXXXXXXXXX no class found for {}".format(node_name))
            return None

    def handle_whitespace(self, node, status, new_parent_attrs, new_parent_labels):
        return []

    def parse_string(self, node, status, parent_attrs, parent_labels):
        """Gets the string  within an opening and a closing tag:
        <title>some string</title>
        ... and stores it as an attribute: object.string
        This method can be overriden to parse the string and create
        individual token objects instead.
        returns an array containing the object
        """
        # either get the string directly found in the tag
        # or within all children
        string = getattr(node, 'string', '') or getattr(node, 'text', '')
        # remove all leading and trailing whitespace
        self.attrs['string'] = string.strip()

    def parse_comment(self, node, status):
        comment_instance = comment(
            node=node, 
            node_name=node.name,
            status=status
        )
        return [comment_instance]

    def _change_base_class(self):
        """During object initialization we will change the parents (base classes)
        of the object dynamically.
        Nested classes like A.B.C do not inherit from each other and there is no way
        to make it work:

        class A(tei):
            class B(A, tei):
                class C(B, A, special_class, tei):
                    ...
        
        Will already fail on compile time.

        To make an object inherit from the classes it is nested in, we need to change
        its base classes, i.e. object.__class__.__bases__. This can be achieved on
        runtime.

        So if we declare our classes like this:
        
        class A(tei):
            class B(tei):
                class C(special_class, tei):
                    pass
        
        the resulting objects will behave as if the classes were defined like this:

        class A(tei):
            class B(A, tei)
                class C(special_class, A.B, A, tei)

        
        Why would we want to have nested classes? Isn't this bad practice?
        ------------------------------------------------------------------

        When parsing a TEI document, we do not treat similar elements
        always the same way.
        For example, a paragraph element <p> can occur almost everywhere
        inside the <text> or <body> elements. 
        It also appears in the <teiHeader>, but there its content
        needs to be treated differently. We could define a general class p which has
        a lot of ifs to decide how to behave in which situation. To avoid this, we
        rather define a class

            text.body.p

        which is a completely different class than

            teiheader.p

        At the same time, we want text.body.p to use some of the specialised methods
        defined in the class text.body, for example.

        The resulting code will be declarative, easy to read and maintain.

        """
        #------------ start changing base class of object --------------
        # get the qualified name of our class without the module name,
        # e.g. 'A.B.C', not '__main__.A.B.C' and not just 'A'
        # NOTE: only works with Python 3.3 and onwards!
        fqn = self.__class__.__qualname__
        class_names = fqn.split('.')

        # get the list of classes we already inherit from
        result_classes = list(self.__class__.__bases__)
        if CHANGE_BASE_CLASS: print("### starting with result_classes: {}".format(list(self.__class__.__bases__)))

        # from the fully qualified class (e.g. A.B.C) we want the given class to
        # inherit from A.B, A (in this order). We put this classes in front of
        # the already present classes (thus insert not append)
        for i in range(1, len(class_names)):
            cls_name = '.'.join(class_names[0:i])
            try:
                cls = eval(cls_name)
                if cls not in result_classes:
                    result_classes.insert(0,cls)
            except AttributeError:
                # class is not explicitly defined (only as virtual metaclass)
                pass
    
        # change the __class__.__bases__ of our newly created object
        # which now is of Classes A.B.C, A.B, A and any other classes it inherits 
        if CHANGE_BASE_CLASS: print("### bases before-> {}".format(self.__class__.__bases__))
        if CHANGE_BASE_CLASS: print("### result_classes-> {}".format(result_classes))
        self.__class__.__bases__ = tuple(result_classes)
        #------------ finish changing base class of object --------------


    def __init__(
            self, node=None, node_name=None, status=None,
            attrs=None, parent_attrs=None, parent_labels=None,
            string=None
        ):
        """ tei class is the parent class for all tei elements.
            node: the current node in the tei-xml file
            node_name: name of the current element (e.g. p, lb, teiheader)
            status: during parsing, certain events like handShift needs to be registered
                  these events affect all next elements
            attrs: attributes inside an element
            parent_attrs: attributes that are passed to all children elements
                Attributes of children might overwrite them again, much like in css.
            string: the actual value (or payload) of an element. Header information
                is not getting interpreted any further, whereas the text itself
                is interpreted (by the parse_string function)
            parent_labels: labels that are passed to all children elements,
                much like parent_attrs.
          
        """

        # make nested classes inherit from each other
        # by dynamically changing their base classes
        self._change_base_class()

        self.xml_node = node

        # set the name of the node, in case it was not provided
        if node_name is None:
            setattr(self, 'node_name', self.__class__.__name__)
        else:
            setattr(self, 'node_name', node_name)

        # add the name of the node as the default label
        self.labels = set()
        self.labels.add(node_name)


        self.child_parent_relation= None

        # If there are any parent attributes, merge them into
        # the attributes of the current child object.
        # make sure the parent attributes are overwritten if a child
        # has the same attributes.
        self.attrs = defaultdict()
        if attrs is None:
            attrs = {}
        if parent_attrs is None:
            parent_attrs = {}
        self.attrs = {**self.attrs, **parent_attrs, **attrs}

        # collect all the attributes we stored in self.attrs 
        # and merge them with node attributes
        node_attrs = getattr(node, 'attrs', {})
        self.attrs = { **self.attrs, **node_attrs}

        # collect all attributes we stored in status['attrs']
        # including the primary key (idno) of the document
        if status is not None:
            for key in status['attrs']:
                self.attrs[key] = status['attrs'][key]

        if parent_labels is None:
            parent_labels = []

        # add delSpan label if we have such a status
        if status.get('delSpan'):
            self.labels.add('delSpan')
            # add delSpan attribute with anchor id
            self.attrs['delSpan'] = status.get('delSpan')

        # add any incoming parent labels to self.labels
        # and make a copy into new_parent_labels
        new_parent_labels = []
        for parent_label in parent_labels:
            self.labels.add(parent_label)
            new_parent_labels.append(parent_label)

        # new_parent_labels now contains all the labels that need to be
        # passed to all children objects
        if getattr(self, 'pass_label_to_children', False):
            new_parent_labels.append(self.pass_label_to_children)

        self.child_instances = []

        # Parse the string of the element itself
        if getattr(self, 'isleaf', False):
            # the node is declared as a leaf, do not parse the node in more detail
            # just extract the text
            self.parse_string(node, status, parent_attrs, parent_labels)
        elif node is None or node.name is None:
            # we encountered just text, which has no node.name and no children either.
            pass
        else:
            # Walk through children of the current node
            new_parent_attrs = {}
            if getattr(self, 'pass_attributes_to_children', False):
                new_parent_attrs = self._get_node_attributes()
                # do not inherit xml:id attribute to children
                if 'xml:id' in new_parent_attrs:
                    new_parent_attrs.pop('xml:id')
            new_parent_attrs = {**parent_attrs, **new_parent_attrs}
        
            for child_node in node.children:
                """- walk through the children of the current node
                - find a corresponding class, either directly or indirectly
                - create an instance of it
                - append the child_instances to the object.child_instances attribute
                """

                if WALK_CHILD_NODES: print("--------WALK_CHILDREN_OF {}-------".format(
                        node.name        
                    )
                )
                if WALK_CHILD_NODES: print(child_node)
                if WALK_CHILD_NODES: print("-------------------------------")

                if child_node.isspace is not None and child_node.isspace():
                    # we did not find any characters, just whitespace
                    child_instances = self.handle_whitespace(
                        child_node, status, new_parent_attrs, new_parent_labels
                    )
                    self.child_instances += child_instances
                    continue

                elif child_node.name is None:
                    # we encountered TEXT or COMMENT, which is not enclosed in tags 
                    # but occurs after a tag (e.g. after a linebreak <lb>)
                    if isinstance(child_node, Comment):
                        # COMMENT
                        continue
                        #self.child_instances += self.parse_comment(child_node, status)
                    else:
                        # we need to parse the TEXT, which means: 
                        # we tokenize it and create token objects
                        child_instances = self.parse_string(
                            child_node, status, new_parent_attrs, new_parent_labels
                        )
                        if child_instances is not None:
                            self.child_instances += child_instances

                else:
                    #
                    # we encountered an child-node (child-element)
                    # find class for child_node
                    targetClass = self.get_class_for_node(child_node)
                    if WALK_CHILD_NODES: print ("got class: {}".format(targetClass))

                    # if we do not find a matching class,
                    # we will ignore the entire element and all its children
                    if targetClass is None:
                        if WALK_CHILD_NODES: print("## node ignored, skipping...")
                        continue
                    else:
                        # we found a matching class: create an instance of it and
                        # append it to self.child_instances
                        if WALK_CHILD_NODES: print("targetClass ----> {} ".format(targetClass))
                            
                        instance = targetClass(
                            node=child_node, node_name=child_node.name,
                            status=status, parent_attrs=new_parent_attrs,
                            parent_labels=new_parent_labels
                        )

                        # if class has the attribute unpack_elements=True
                        # we want to directly append the children of the children
                        # to the current instance.
                        if instance.unpack_elements:
                            if instance.create_empty_token_nodes \
                                and len(instance.child_instances) == 0:
                                token_obj = Token(
                                    node=child_node, node_name='token',
                                    status=status, 
                                    attrs={"string": "", "whitespace": ""},
                                    parent_labels=new_parent_labels+[instance.node_name]
                                ) 
                                self.child_instances.append(token_obj)
                            
                            if getattr(instance, 'parse_attrs', False):
                                instance.parse_attrs(
                                    node=child_node, 
                                    status=status,     
                                    parent_labels=new_parent_labels,
                                    parent_attrs=new_parent_attrs,
                                )

                            for child_instance in instance.child_instances: 
                                self.child_instances.append(
                                    child_instance
                                )
                        else:
                            self.child_instances.append(instance)

        self.create_parent_child_relationships(status)
        self._group_similar_siblings(status)


    def _get_node_attributes(self):
        """returns all attributes of an XML element as a dict
        """
        attrs = {}
        for attr in self.xml_node.attrs:
            attrs[attr] = self.xml_node.attrs[attr]
        return attrs


    def create_parent_child_relationships(self, status):
        """This is the default relation between parent and child objects: CONTAINS
        Certain classes (like p) might implement their own, specific relations.
        If you just want to override the default, add parent_child_relation='WHATEVER'
        to your class definition
        """
        self.relationships=[]
        label = getattr(self, 'parent_child_relation', 'CONTAINS')
        for i, child_instance in enumerate(self.child_instances):
            if PARENT_CHILD_RELATION: print(self.__class__.__name__ + " CONTAINS " + child_instance.__class__.__name__)
            self.relationships.append({
                "from": self,
                "to": child_instance,
                "label": label,
            })


    def _group_similar_siblings(self, status):
        """Creates additional relationships between similar siblings
        e.g. NEXT_LB, NEXT_TOKEN
        Also creates relationships that point to the first and last element:
        FIRST_LB, LAST_LB
        Classes must be declared this way:
        class body(tei, group_similar_siblings=['p'])
        """
        if getattr(self, 'group_similar_siblings', None) is None:
            return

        prev_instance = {}
        counter = defaultdict(int)
        for instance in [
            child_instance for child_instance in self.child_instances\
                if child_instance.node_name in self.group_similar_siblings
        ]:
            counter[instance.node_name] += 1
            # FIRST appearance
            if counter[instance.node_name] == 1:
                self.relationships.append({
                    "from" : self, 
                    "to"   : instance,
                    "label": "FIRST_"+instance.node_name.upper()
                })

            if prev_instance.get(instance.node_name):
                self.relationships.append({
                    "from" : prev_instance[instance.node_name],
                    "to"   : instance,
                    "label": "NEXT_"+instance.node_name.upper()
                })

            prev_instance[instance.node_name] = instance
            
        # LAST
        for node_name in self.group_similar_siblings:
            if prev_instance.get(node_name):
                self.relationships.append({
                    "from" : self, 
                    "to"   : prev_instance[node_name],
                    "label": "LAST_"+ node_name.upper()
                })

    def _update_node(self, node, labels, attrs):
        for label in labels:
            node.add_label(label)
        for attr in attrs:
            node[attr] = attrs[attr]
        

    def _node_for_obj(self, obj, tx, node_cache):
        attrs = getattr(obj, 'attrs', {})
        #matcher = NodeMatcher(tx.graph)

        if not getattr(obj, 'node', False):
            if CREATE_NODE: print("creating Node: {}".format(obj.node_name))
            if CREATE_NODE: print("               {}".format(attrs.get('string')))
            if getattr(obj, 'parse_attrs', False):
                obj.parse_attrs()

            labels = list(obj.labels)
            if CREATE_NODE: print("LABELS: {}".format(labels))

            # if we got a node with an xml:id we first want to check whether we
            # already have an empty node to update
            xml_id = attrs.get('xml:id')
            filename = attrs.get('filename')
            if xml_id and filename:
                node=None
                if filename in node_cache and xml_id in node_cache[filename]:
                    # search in cache
                    node = node_cache[filename][xml_id]
                    self._update_node(node, labels, attrs)
                else:
                    # search in database
                    search_attrs = {
                        'filename' : filename,
                        'xml_id'   : xml_id
                    }
                    cypher = """
                        MATCH (n) WHERE n.filename = $filename
                        AND             n.`xml:id` = $xml_id
                        RETURN n
                    """
                    cursor = tx.graph.run(cypher, parameters=search_attrs)
                    node = None
                    if cursor.forward():
                        node=cursor.current
                    #node = matcher.match(**search_attrs).first()

                    if (node):
                        # update existing node in database: add labels and attrs
                        self._update_node(node, labels, attrs)
                        tx.graph.push(node)
                    else:
                        # node is neither in cache nor in database.
                        # create a new empty node
                        node = Node(
                            *labels,
                            **attrs
                        )
                        tx.create(node)

                        # store node in node_cache
                        if not filename in node_cache:
                            node_cache[filename] = {}
                        node_cache[filename][xml_id] = node


                #if node:
                #    # we found a node, either in the current transaction
                #    # or in the database
                #    for label in labels:
                #        node.add_label(label)
                #    for attr in attrs:
                #        node[attr] = attrs[attr]

                #    if tx.exists(node):
                #        # update node in transaction:
                #        # we need to delete and create it again
                #        #tx.delete(node)
                #        #tx.create(node)
                #        pass
                #    else:
                #        # existing node in database:
                #        # update database
                #        tx.graph.push(node)
                #else:
                #    # create a new node object
                #    node = Node(
                #        *labels,
                #        **attrs
                #    )
                #    tx.create(node)
                #if not filename in node_cache:
                #    node_cache[filename] = {}
                ## store node in cache
                #node_cache[filename][xml_id] = node

            # we do not have a node with xml:id attribute
            else:
                # create node
                node = Node(
                    *labels,
                    **attrs
                )
                tx.create(node)
                # we do not need to store this node in the node_cache, since
                # it has no xml:id and therefore can't be referenced.

            """ 
            if the element links to another node:
            1. check if that node already exists (xml:id, filename)
               a) check it in the current status cache (because node is not saved yet)
               b) check in the current graph
               -> exists: create relationship with attribute name as label
               -> does not exist: create placeholder node and relationship
            """ 
            
            obj.node = node
        return obj.node


    def node_for_labels_attrs(self, labels, attrs, tx):
        xml_id = attrs.get('xml:id')
        filename = attrs.get('filename')
        #matcher = NodeMatcher(tx.graph)
        if xml_id and filename:
            node=None
            if filename in node_cache and xml_id in node_cache[filename]:
                # search in cache
                node = node_cache[filename][xml_id]
            else:
                # search in database
                search_attrs = {
                    'filename' : filename,
                    'xml_id'   : xml_id
                }
                cypher = """
                    MATCH (n) WHERE n.filename = $filename
                    AND             n.`xml:id` = $xml_id
                    RETURN n
                """
                cursor = tx.graph.run(cypher, parameters=search_attrs)
                if cursor.forward():
                    node=cursor.current
                #node = matcher.match(**search_attrs).first()

            if node:
                # we found a node, either in the current transaction
                # or in the database
                for label in labels:
                    node.add_label(label)
                for attr in attrs:
                    node[attr] = attrs[attr]

                if tx.exists(node):
                    # update node in transaction:
                    # we need to delet it and create it again
                    #tx.delete(node)
                    #tx.create(node)
                    pass
                else:
                    # existing node in database:
                    # update database
                    tx.graph.push(node)
            else:
                # create a new node object
                node = Node(
                    *labels,
                    **attrs
                )
                tx.create(node)
            if not filename in node_cache:
                node_cache[filename] = {}
            # store node in cache
            node_cache[filename][xml_id] = node
        else:
            # create node
            node = Node(
                *labels,
                **attrs
            )
            tx.create(node)
            # we do not need to store this node in the node_cache, since
            # it has no xml:id and therefore can't be referenced.


    def save(self, tx, node_cache=None):
        """Recursively walk through all child_instances generated and 
        create nodes and relationships in the graph database
        """
        if node_cache is None:
            node_cache = {}

        # create node for the current element
        # adding element attributes, e.g. <p attr1="one" attr2="two">
        # as attributes to the node itself

        # create all nodes (if they do not exist yet) and relationships
        for relationship in self.relationships:
            from_node = self._node_for_obj(relationship['from'], tx, node_cache)
            to_node   = self._node_for_obj(relationship['to'], tx, node_cache)
            attr = relationship.get('attr', {})
            rs = Relationship(
                from_node,
                relationship['label'],
                to_node,
                **attr
            )
            tx.create(rs)

        for child_instance in self.child_instances:
            child_instance.save(tx, node_cache)


class include_all_elements():
    def get_class_for_node(self, node):
        """Inside a given section we want to follow every element, 
        even the ones we have no class definition for.
        We therefore have to instantly create a new class using type()
        """
        cls = super().get_class_for_node(node)
        if cls is None and node.name is not None:
            node_name = node.name

            if GET_CLASS_FOR_NODE: print("...creating new class for node: {}".format(node_name))
            parent_classes = [parent.name for parent in node.parents][:-2]
            parent_classes.reverse()
            parent_classes.append(node_name)

            # add an underscore to the class name 
            # in case it is a reserved python word
            # --> reserved python words cannot be used as a class name
            for i, class_name in enumerate(parent_classes):
                if class_name in keyword.kwlist:
                    parent_classes[i] = class_name + '_'

            class_name = ".".join(parent_classes)
            class_name = class_name.lower()
            if GET_CLASS_FOR_NODE: print("...creating new class: {}".format(class_name))

            return type(class_name, (tei, object), {})
        else:
            return cls


class ParseString(include_all_elements, tei):
    def parse_string(self, node, status, parent_attrs, parent_labels):
        """This method is parsing the string between elements by dividing
        it into Token objects
        - all token objects are stored in a variable token_objs which is then
            returned.
        It detects if a token represents a hyphened word fragment by either:
        - checking if it is the token ends with a - or a non-breaking hyphen
        - looking behind if it is the first word after a line break and the
        - line break is of type="hyph"
        """
        line = getattr(node, 'string', '')
        if line is None:
            return []

        line = str(line)

        # remove newlines and whitespace / tabs that are after a newline,
        # since they belong to XML formatting, not to actual wanted text
        line = re.sub(r'[\n\r]+([\t\s]*)', '', line, flags=re.MULTILINE)

        # we did not see anything other than tabs and newlines...
        if line == '':
            return []
        # (but we are still interested in space ;-)

        # finally do the tokenization using spacy's nlp
        tokens = nlp(line)

        token_objs = []
        # Walk through text-tokens and create Token objects
        if PARSE_STRING: print("TOKENS: «{}»".format(tokens))
        for i, token in enumerate(tokens):
            is_hyphened = False
            if any(
                token.text.endswith(s) for s\
                in ['-', '\N{NOT SIGN}', '\N{NON-BREAKING HYPHEN}']
            ):
                is_hyphened = True

            attrs = {
                'is_punct'        : token.is_punct,
                'pos'             : token.pos_,
                'tag'             : token.tag_,
                'is_hyphened'     : is_hyphened,
                'string'          : token.text,
                'whitespace'      : token.whitespace_,  # token is followed by a whitespace
                # from handshift
                'medium': status.get('handshift').get('medium'),
                'script': status.get('handshift').get('script'),
            }

            # Merge parent_attrs into attrs
            # but let them to be overwritten by the child elements
            if parent_attrs:
                attrs = {**parent_attrs, **attrs}

            # Create a Token object for every token
            token_obj = Token(
                node=node, node_name='token',
                status=status, attrs=attrs, parent_labels=parent_labels
            ) 

            # add it to our collection of token objects
            token_objs.append(token_obj)

        # these token objects get appended to .child_instances
        return token_objs

    handle_whitespace=parse_string


class ConnectSiblings:
    """This class defines how siblings, i.e. elements on the same hierarchical level,
    are connected with
    - each other
    - between the parent and the first and last children

    Default:
    parent-FIRST->sibling-NEXT->sibling<-LAST-parent

    Classes that inherit from this class can define attributes (see below)
    to define specific behaviour.
    """
    def __init_subclass__(cls, 
        parent_to_first_sibling_relation='FIRST',
        first_sibling_to_parent_relation=None,
        parent_to_last_sibling_relation='LAST',
        last_sibling_to_parent_relation=None,
        previous_next_sibling_relation='NEXT', 
        previous_next_relation_attr=None,
        next_previous_sibling_relation=None, 
        next_previous_relation_attr=None, 
        **kwargs
    ):
        super().__init_subclass__(**kwargs)
        cls.parent_to_first_sibling_relation = parent_to_first_sibling_relation
        cls.parent_to_last_sibling_relation = parent_to_last_sibling_relation
        cls.last_sibling_to_parent_relation = last_sibling_to_parent_relation
        cls.first_sibling_to_parent_relation = first_sibling_to_parent_relation

        cls.previous_next_sibling_relation = previous_next_sibling_relation
        if previous_next_relation_attr is None:
            previous_next_relation_attr = {}
        cls.previous_next_relation_attr = previous_next_relation_attr

        cls.next_previous_sibling_relation = next_previous_sibling_relation
        if next_previous_relation_attr is None:
            next_previous_relation_attr = {}
        cls.next_previous_relation_attr = next_previous_relation_attr

    def create_parent_child_relationships(self, status):
        """Replaces the usual parent-child relationships (CONTAINS)
        by pointing from the parent to the FIRST sibling, 
        then connecting to the NEXT sibling
        and finally pointing from the parent to the LAST sibling
        """
        prev_child_instance = None
        self.relationships=[]

        # skip child instances that will not be stored
        child_instances = list(
            child for child in self.child_instances 
            if not child.do_not_store_this_node
        )

        for i, child_instance in enumerate( child_instances ):
            # First sibling
            if i == 0:
                if self.parent_to_first_sibling_relation:
                    self.relationships.append({
                        "from": self,
                        "to": child_instance,
                        "label": self.parent_to_first_sibling_relation,
                    })
                if self.first_sibling_to_parent_relation:
                    self.relationships.append({
                        "from": child_instance,
                        "to": self,
                        "label": self.first_sibling_to_parent_relation,
                    })

            if prev_child_instance is not None:
                if self.previous_next_sibling_relation:
                    self.relationships.append({
                        "from": prev_child_instance,
                        "to": child_instance,
                        "label": self.previous_next_sibling_relation,
                        "attr": self.previous_next_relation_attr,
                    })
                if self.next_previous_sibling_relation:
                    self.relationships.append({
                        "from": child_instance,
                        "to": prev_child_instance,
                        "label": self.next_previous_sibling_relation,
                        "attr": self.next_previous_relation_attr,
                    })


            # last sibling
            if i == len(child_instances)-1:
                if self.parent_to_last_sibling_relation:
                    self.relationships.append({
                        "from": self,
                        "to": child_instance,
                        "label": self.parent_to_last_sibling_relation,
                    })
                if self.last_sibling_to_parent_relation:
                    self.relationships.append({
                        "from": child_instance,
                        "to": self,
                        "label": self.last_sibling_to_parent_relation,
                    })
            prev_child_instance = child_instance


# TEIHEADER section
# only the elements listed here are being parsed and
# added to the database. Strings in an element are,
# as opposed to the BODY section, not being tokenized
# (split into words etc.). 
class teiheader(tei):
    class filedesc(include_all_elements, tei):
        pass

    class encodingdesc(tei):
        class editorialdecl(tei, isleaf=True):
            pass

    class profiledesc(include_all_elements, tei):
        pass

    class revisiondesc(include_all_elements, tei):
        pass


# FACSIMILE section
# the facsimilies section defines zones (coordinates) on the image itself
# every zone contains an xml:id which is used in paragraphs to refer to: <p facs="#facs_1_r3">
class facsimile(tei, cache_object=True):
    class surface(tei):
        class graphic(tei):
            pass
        class zone(tei, cache_object=True):
            pass
    

# TEXT section
# <text> is actually holding the real content in every TEI document
class text(include_all_elements, tei):
    def get_class_for_node(self, node):
        """Inside the <text> section we want to follow every element, 
        even the ones we have no class definition for.
        We therefore have to instantly create a new class using type()
        """
        cls = super().get_class_for_node(node)
        if cls is None and node.name is not None:
            node_name = node.name

            if VERBOSE > 1: print("...creating new class for node: {}".format(node_name))
            parent_classes = [parent.name for parent in node.parents][:-2]
            parent_classes.reverse()
            parent_classes.append(node_name)

            # add an underscore to the class name 
            # in case it is a reserved python word
            # --> reserved python words cannot be used as a class name
            for i, class_name in enumerate(parent_classes):
                if class_name in keyword.kwlist:
                    parent_classes[i] = class_name + '_'


            if VERBOSE > 1: print("...creating: {}".format(parent_classes))
            class_name = ".".join(parent_classes)
            if VERBOSE > 1: print("...creating new class: {}".format(class_name))

            return type(class_name, (tei, object), {})
        else:
            return cls

    class front(tei):
        pass

    class body(
        ConnectSiblings,
        tei,
        parent_to_first_sibling_relation='FIRST_ELEMENT',
        previous_next_sibling_relation='NEXT_ELEMENT',
    ):

        class p(
            ParseString,
            ConnectSiblings,
            group_similar_siblings=['lb'],
            parent_to_first_sibling_relation='NEXT',
            parent_to_last_sibling_relation='LAST',
            previous_next_sibling_relation='NEXT'
        ):
            """ handles all paragraphs tags <p> within the <body> tag
            """
            def __init__(self, 
                node=None, node_name=None, status=None,
                parent_attrs=None, parent_labels=None
            ):
                # add paragraph_id to status attrs,
                # as we add the paragraph_id to all nodes
                if node.get("xml:id"):
                    status["attrs"]["paragraph_id"] = node.get("xml:id")
                super().__init__(
                    node=node, node_name=node_name, status=status,
                    parent_attrs=parent_attrs, parent_labels=parent_labels     
                )
                # remove paragraph_id again when we leave
                # this paragraph
                if "paragraph_id" in status["attrs"]:
                    status["attrs"].pop("paragraph_id")
                    
        class pb(ParseString):
            pass
        class fw(ParseString):
            pass
        class undo(ParseString):
            pass
    class back(ParseString):
        pass

class comment(tei):
    pass


class Token(
    ConnectSiblings, 
    tei,
):
    pass


class Word(
    ConnectSiblings,
    tei,
):
    pass


class handshift(tei, do_not_store_this_node=True):
    def __init__(
        self, node=None, node_name=None, status=None, 
        attrs=None, parent_attrs=None, parent_labels=None,
        string=None
    ):

        # only overwrite the latest handshifts, keep everything else.
        if node.get('new'):
            status['handshift']['new'] = node.get('new')
        if node.get('medium'):
            status['handshift']['medium'] = node.get('medium')
        if node.get('script'):
            status['handshift']['script'] = node.get('script')

        super().__init__(node=node, node_name=node_name, status=status)


class delspan(tei):
    """a delSpan element marks all following elements to be deleted,
    until an anchor element appears that contains the ID of the spanTo attribute.
    Albeit currently not needed, we still store this node in the graph.
    """
    def __init__(
        self, node=None, node_name=None, status=None, 
        attrs=None, parent_attrs=None, parent_labels=None,
        string=None
    ):

        if node.get('spanTo'):
            spanTo = node.get('spanTo')
            # find first occurence of the anchor tag #
            # and store this
            spanTo = spanTo[spanTo.find('#')+1:]
            status['delSpan'] = spanTo

        super().__init__(node=node, node_name=node_name, status=status)


class anchor(tei):
    """a delSpan element always points to an anchor object. Because of this, we need
    to alter the status back to normal when we encounter an anchor object.
    """
    def __init__(
        self, node=None, node_name=None, status=None, 
        attrs=None, parent_attrs=None, parent_labels=None,
        string=None
    ):

        if status.get('delSpan'):
            if status.get('delSpan') == node.get('xml:id'):
                status.pop('delSpan')

        super().__init__(node=node, node_name=node_name, status=status)


class rs(ParseString, 
    unpack_elements=True,
    pass_label_to_children="rs", 
    create_empty_token_nodes=True,
    pass_attributes_to_children=True
):
    """Reference String
    see https://tei-c.org/release/doc/tei-p5-doc/en/html/ref-rs.html
    """
    pass


class unclear(ParseString, 
    unpack_elements=True,
    pass_label_to_children="unclear", 
    pass_attributes_to_children=True
):
    """ Elements <unclear> will not create a node in the graph,
    but they will pass their attributes to the children they contain
    and add an additional label :unclear
    """
    pass

    def parse_attrs(self, node, status, parent_labels, parent_attrs):
        """handles the extent attribute, e.g.
        adds an additional node with ¿¿¿
        depending on the number in extent="3 Graphen"
        """
        if 'extent' in self.attrs:
            token_obj = Token(
                node=node,
                node_name='token',
                status=status, 
                parent_labels=parent_labels,
                attrs=parent_attrs
            ) 
            match = re.search(
                r'(?P<number_of_graphs>\d+)\s+Graph',
                self.attrs['extent']
            )
            if match: 
                string = ""
                try:
                    string="¿"*int(
                        match.groupdict()['number_of_graphs']
                    )
                    token_obj.attrs["string"] = string
                    token_obj.attrs["whitespace"] = ""
                except Exception:
                    pass
 
            self.child_instances.insert(0,token_obj)


class add(
    ParseString,
    unpack_elements=True,
    pass_label_to_children="add",
    pass_attributes_to_children=True,
):
    """During parsing, this class should add the label "add"
    to every child element and all attributes should be passed to its children.
    """

class del_(
    ParseString,
    unpack_elements=True,
    pass_label_to_children="del",
    pass_attributes_to_children=True,
):
    """During parsing, this class should add the label "del"
    to every child element and all attributes should be passed to its children.
    """

class subst(
    ParseString,
    unpack_elements=True,
    pass_label_to_children="subst",
    pass_attributes_to_children=True,
):

    """
    Substitutions add some interpretation to the text,
    e.g. they suggest that a correction is done in one
    single pass. The del and add elements inside a subst
    are therefore grouped together.
    """

class metamark(
    ParseString,
    unpack_elements=True,
    pass_label_to_children="metamark",
    pass_attributes_to_children=True,
):
    pass

class hi(
    ParseString,
    unpack_elements=True,
    pass_label_to_children="hi",
    pass_attributes_to_children=True,
):
    pass


class sic(
    ParseString,
    unpack_elements=True,
    pass_label_to_children="sic",
    create_empty_token_nodes=True,
    pass_attributes_to_children=True
):
    pass

class corr(
    ParseString,
    unpack_elements=True,
    pass_label_to_children="corr",
    pass_attributes_to_children=True
):
    pass

class choice(
    ParseString,
    unpack_elements=True,
    pass_label_to_children="choice",
):
    pass



import sys
import os
import re
from py2neo import Graph, Node, Relationship, NodeMatcher, RelationshipMatcher
from tei2neo import parse, GraphUtils


def my_sort(coll):
    def my_coll_sort(val):
        match = re.search(r'\d+_Ms_{}_(?P<page>\d+)'.format(coll), val)
        if match:
            return int(match.groupdict()['page'])
        else:
            return 0
    return my_coll_sort


def my_filter(coll):
    def my_coll_filter(val):
        match = re.search(r'^\d+_Ms_{}.*?xml$'.format(coll), val)
        return match
    return my_coll_filter


def get_graph():
    host = input('hostname (localhost):') or 'localhost'
    port = input('port (7687):') or '7687'
    user = input('port (neo4j):') or 'neo4j'
    import getpass
    pw = os.environ.get('SEMPER_NEO4J_PASSWORD')
    password = getpass.getpass('password ({}):'.format(pw)) or pw

    graph = Graph(
        host = host, 
        port = port,
        user = user,
        password = password,
    )
    return graph


def import_file(filepath, graph=None):
    if graph is None:
        graph = get_graph()
    ut = GraphUtils(graph)

    filepath = os.path.abspath(filepath)
    filename = os.path.basename(filepath)

    ut.delete_graph_for_file(filename)

    # import
    doc, status, soup = parse(filename=filepath)
    tx = graph.begin()
    doc.save(tx)
    tx.commit()

    # create the relationships within the document
    ut.link_inner_relationships(filename)

    # connect the document to existing categories
    ut.connect_to_categories(filename)

    # get all paragraphs in a file
    paras = ut.paragraphs_for_filename(filename)

    # create the unhyphened tokens
    for para in paras:
        tokens = ut.tokens_in_paragraph(para)
        ut.create_unhyphenated(tokens)


def get_files_for_path(path):
    filenames = []
    # get the last directory name as an indicator for the collection
    collection = os.path.basename(os.path.normpath(path))
    for root, dirs, files in os.walk(path):
        for filename in sorted(
                filter(my_filter(coll=collection), files),
                key=my_sort(coll=collection)
        ):
            filepath = os.path.join(root, filename)
            filenames.append({
                "filepath": filepath,
                "filename": filename,
            })

    return filenames


def import_files_from_path(path, graph=None):
    """Imports all xml files from a given directory.
    These files must fit the pattern and are ordered by their page number
    before they are imorted.
    """
    if os.path.isfile(path):
        import_file(filenpath=path, graph=graph)

    if graph is None:
        graph = get_graph()
    ut = GraphUtils(graph)

    files = get_files_for_path(path)
    first_filename = files[0]
    last_filename = files[:-1]
    for f in files:
        import_file(f["filepath"], graph)

    just_filenames = [f["filename"] for f in files]
    ut.connect_teis(*just_filenames)


if __name__ == '__main__':
    import_files_from_path(sys.argv[1])



import os
import sys

if sys.version_info < (3,5):
    sys.exit('Sorry, Python < 3.5 is not supported')

from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='tei2neo',
    version= '0.2.2',
    description='TEI (Text Encoding Initiative) parser to extract information and store it in Neo4j database',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://sissource.ethz.ch/sis/semper-tei',
    author='Swen Vermeul • ID SIS • ETH Zürich',
    author_email='swen@ethz.ch',
    license='BSD',
    packages=[
        'tei2neo',
    ],
    install_requires=[
        'pytest',
        'py2neo',
        'bs4',
        'lxml',
        'spacy',
        'requests',
    ],
    python_requires=">=3.5",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)




def classes_for_token(token):
    """This function returns a set of classes for a given token.
    Depending on its script, medium and rend attribute,
    appropriate css-class names are returned
    """

    medium2style = {
        #"Blei"   : "color: gray",
        #"STinte" : "color: black",
        #"BlTinte": "color: blue",
        #"BrTinte": "color: brown",
    }
    medium2class = {
        #"Blei"    : "tei-Blei",
        #"STinte"  : "tei-STinte",
        #"BlTinte" : "tei-BlTinte",
        #"BrTinte" : "tei-BrTinte",
    }
    rend2style = {
        "unterstr": "text-decoration: underline",
        "durchstr": "text-decoration: line-through",
        "überschr": "vertical-align: super",
        "hoch"    : "vertical-align: super",
        #"durchstr BlTinte": "text-decoration: line-through; text-decoration-color: blue",
        #"durchstr BrTinte": "text-decoration: line-through; text-decoration-color: brown",
        #"durchstr STinte" : "text-decoration: line-through; text-decoration-color: black",
        #"Blei"    : "color: gray",
        #"BrTinte" : "color: brown",
    }
    rend2class = {
        "unterstr": "tei-underline",
        "durchstr": "tei-line-through",
        "überschr": "tei-super",
        "hoch"    : "tei-super",
        #"durchstr BlTinte": "tei-line-through tei-line-through-blue",
        #"durchstr BrTinte": "tei-line-through tei-line-through-brown",
        #"durchstr STinte" : "tei-line-through",
        #"Blei"    : "tei-Blei",
        #"BrTinte" : "tei-BrTinte",
    }
    script2class = {
        #"Kurrent"   : "tei-Kurrent",
        #"Latein"    : "tei-Latein",
        #"Griechisch": "tei-Griechisch"
    }
    
    classes = []
    for script in script2class.keys():
        if script in token.get('script',''):
            classes.append(script2class[script])
    for medium in medium2class.keys():
        if medium in token.get('medium',''):
            classes.append(medium2class[medium])
    for rend in rend2class.keys():
        if rend in token.get('rend',''):
            classes.append(rend2class[rend])

    # add the labels themselves as classes
    classes += ["tei-{}".format(label) for label in token.labels]

    return set(classes)
